
/* (stolen from fifi, http://puska.keltti.jyu.fi/fifi) */



/* list.c - operations with (sorted) lists
 */

/* $Id: list.c,v 1.2 2002/08/18 09:52:46 engerim Exp $ */

/*
 * $Log: list.c,v $
 * Revision 1.2  2002/08/18 09:52:46  engerim
 * Added files of jv
 *
 * Revision 1.7  1999/06/14 14:22:29  liff
 * i did make indent
 *
 * Revision 1.6  1999/06/14 10:23:23  liff
 * Added add_list_to_list.
 * I didn't need it tho :)
 *
 * Revision 1.5  1999/05/14 07:14:45  pht
 *
 * you can now do 'make indent' :>
 *
 * Revision 1.4  1999/05/13 19:32:26  steven
 * added functions to add to a sorted list (list.c list.h)
 * added functions to handle time hooks (hook.c hook.h)
 * modified core.c to test
 * made #define for channel and server to join :p
 * (because I'm lazy to always find it and change it)
 *
 * Revision 1.3  1999/05/11 15:52:41  steven
 * changed *.h so they can be #included several times
 * (with #ifndef and stuff)
 *
 * Revision 1.2  1999/05/07 00:24:48  pht
 *
 * Id/Log-ized.
 *
 */

#include <sys/types.h>
#include "list.h"
#include <string.h>
#include "mem.h"

/* add_to_list - adds a List compatible structure in front of a linked list
 * parameters:
 * first, last - pointers to pointers to the first and last structures
 *               last can be NULL if there is no need for it
 * size - number of bytes to allocate for the structure
 * name - string identifier of the structure
 * id - numeric identifier
 * return value:
 * pointer to newly created structure, or NULL if there is not enough memory
 */

inline static int int_sub(register int a, register int b)
{
    return a - b;
}

inline static int nullcheck_strcmp(const char *a, const char *b)
{
    if (a && b)
	return strcmp(a, b);

    return (unsigned int) a - (unsigned int) b;
}


List *add_to_list(List ** first, List ** last, size_t size, char *name,
		  int id)
{
    register List *list;

    list = my_calloc(1, size);
    if (!list)
	return NULL;

    if (name)
    {
	list->name = my_calloc(1, strlen(name) + 1);
	if (!list->name)
	{
	    my_free(list);
	    return NULL;
	}
	strcpy(list->name, name);
    }
    list->id = id;

    if (!*first)
    {
	if (last)
	    *last = list;
    } else
    {
	list->next = *first;
	(*first)->prev = list;
    }

    *first = list;
    return list;
}

/* add_list_to_list - prepends a list into another list
 * parameters:
 * l1 is prepended to l2
 * return value:
 * pointer to the new and improved list; usually just l1
 */
List *
 add_list_to_list(List * l1, List * l2)
{
    if (!l1)
	return l2;
    else
    {
	List *p;

	for (p = l1; p->next; p = p->next)
	    ;
	p->next = l2;

	return l1;
    }
}

/* add_to_id_sorted_list - adds a List compatinle structure to a sorted list
 * parameters:
 * first, last - pointers to pointers to the first and last structures
 *               last can be NULL if there is no need for it
 * size - number of bytes to allocate for the structure
 * name - string identifier of the structure
 * id - numeric identifier
 * cmp_func() - function to sort with (can be NULL -> 
 *                              sorting by size) 
 * return value:
 * pointer to newly created structure, or NULL if there is not enough memory
 */
List *add_to_id_sorted_list(List ** first, List ** last, size_t size,
			    char *name, int id, int cmp_func(int a, int b))
{
    List *temp_list, *temp_prev = NULL, *p;

    if (!cmp_func)
	cmp_func = int_sub;
    if (!*first)
    {
	p = add_to_list(first, last, size, name, id);
    } else
    {
	temp_list = *first;
	temp_prev = temp_list->prev;
	while (temp_list)
	{
	    if (cmp_func(temp_list->id, id) <= 0)
	    {
		temp_prev = temp_list;
		temp_list = temp_list->next;
	    } else
		break;
	}
	/* temp_list is the struct where cmp_func failed , so it will be p->next
	 * temp_prev is the struct where cmp_func was last succesfull , 
	 * so it will be p->prev;
	 */
	p = my_calloc(1, size);
	if (!p)
	    return NULL;
	if (name)
	{
	    p->name = my_calloc(1, strlen(name) + 1);
	    if (!p->name)
	    {
		my_free(p);
		return NULL;
	    }
	    strcpy(p->name, name);
	}
	p->id = id;
	p->next = temp_list;
	p->prev = temp_prev;
	if (!temp_list)
	{
	    if (last)
		*last = p;
	} else
	{
	    temp_list->prev = p;
	}
	if (!temp_prev)
	{
	    *first = p;
	} else
	{
	    temp_prev->next = p;
	}
    }
    return p;
}

/* add_to_name_sorted_list - adds a List compatinle structure to a sorted list
 * parameters:
 * first, last - pointers to pointers to the first and last structures
 *               last can be NULL if there is no need for it
 * size - number of bytes to allocate for the structure
 * name - string identifier of the structure
 * id - numeric identifier
 * cmp_func() - function to sort with (can be NULL ->
 *                              sorting by ASCII table)
 * return value:
 * pointer to newly created structure, or NULL if there is not enough memory
 */


List *add_to_name_sorted_list(List ** first, List ** last, size_t size,
			      char *name, int id, int cmp_func(const char
							       *a, const char *b))
{
    List *temp_list, *temp_prev = NULL, *p;

    if (!cmp_func)
	cmp_func = nullcheck_strcmp;
    if (!first)
    {
	p = add_to_list(first, last, size, name, id);
    } else
    {
	temp_list = *first;
	temp_prev = temp_list->prev;
	while (temp_list)
	{
	    if (cmp_func(temp_list->name, name) <= 0)
	    {
		temp_prev = temp_list;
		temp_list = temp_list->next;
	    } else
		break;
	}
	/* temp_list is the struct where cmp_func failed , so it will be p->next
	 * temp_prev is the struct where cmp_func was last succesfull , 
	 * so it will be p->prev;
	 */
	p = my_calloc(1, size);
	if (!p)
	    return NULL;
	if (name)
	{
	    p->name = my_calloc(1, strlen(name) + 1);
	    if (!p->name)
	    {
		my_free(p);
		return NULL;
	    }
	    strcpy(p->name, name);
	}
	p->id = id;
	p->next = temp_list;
	p->prev = temp_prev;
	if (!temp_list)
	{
	    if (last)
		*last = p;
	} else
	{
	    temp_list->prev = p;
	}
	if (!temp_prev)
	{
	    *first = p;
	} else
	{
	    temp_prev->next = p;
	}
    }
    return p;
}

/* remove_from_list - removes structure from list
 * params:
 * first, last - see add_to_list
 * p - struct to remove
 */

void remove_from_list(List ** first, List ** last, List * p)
{
    if (!p)
	return;

    if (p->next)
	p->next->prev = p->prev;
    else if (last)
	*last = p->prev;
    if (p->prev)
	p->prev->next = p->next;
    else
	*first = p->next;

    if (p->name)
	my_free(p->name);
    my_free(p);
}


List *find_in_list_by_name(List * first, char *name,
			   int cmp_func(const char *a, const char *b))
{
    register List *p;

    if (!cmp_func)
	cmp_func = nullcheck_strcmp;

    for (p = first; p; p = p->next)
	if (!cmp_func(p->name, name))
	    return p;

    return NULL;
}


List *find_in_list_by_id(List * first, int id,
			 int cmp_func(int a, int b))
{
    register List *p;

    if (!cmp_func)
	cmp_func = int_sub;

    for (p = first; p; p = p->next)
	if (!cmp_func(p->id, id))
	    return p;

    return NULL;
}
